@@ -15,6 +15,7 @@
 #include <string>
 #include <cmath>
 #include <map>
+#include <memory>
 #include <list>
 #include <vector>
 #include "libvisio_utils.h"
@@ -37,144 +38,143 @@ public:
     std::vector<std::list<unsigned> > &documentPageShapeOrders,
     VSDStyles &styles, VSDStencils &stencils
   );
-  virtual ~VSDContentCollector()
-  {
-    if (m_txtxform) delete(m_txtxform);
-  };
 
-  void collectEllipticalArcTo(unsigned id, unsigned level, double x3, double y3, double x2, double y2, double angle, double ecc);
-  void collectForeignData(unsigned level, const librevenge::RVNGBinaryData &binaryData);
-  void collectOLEList(unsigned id, unsigned level);
-  void collectOLEData(unsigned id, unsigned level, const librevenge::RVNGBinaryData &oleData);
-  void collectEllipse(unsigned id, unsigned level, double cx, double cy, double xleft, double yleft, double xtop, double ytop);
+  void collectDocumentTheme(const VSDXTheme *theme) override;
+  void collectEllipticalArcTo(unsigned id, unsigned level, double x3, double y3, double x2, double y2, double angle, double ecc) override;
+  void collectForeignData(unsigned level, const librevenge::RVNGBinaryData &binaryData) override;
+  void collectOLEList(unsigned id, unsigned level) override;
+  void collectOLEData(unsigned id, unsigned level, const librevenge::RVNGBinaryData &oleData) override;
+  void collectEllipse(unsigned id, unsigned level, double cx, double cy, double xleft, double yleft, double xtop, double ytop) override;
   void collectLine(unsigned level, const boost::optional<double> &strokeWidth, const boost::optional<Colour> &c, const boost::optional<unsigned char> &linePattern,
                    const boost::optional<unsigned char> &startMarker, const boost::optional<unsigned char> &endMarker,
-                   const boost::optional<unsigned char> &lineCap, const boost::optional<double> &rounding);
+                   const boost::optional<unsigned char> &lineCap, const boost::optional<double> &rounding,
+                   const boost::optional<long> &qsLineColour, const boost::optional<long> &qsLineMatrix) override;
   void collectFillAndShadow(unsigned level, const boost::optional<Colour> &colourFG, const boost::optional<Colour> &colourBG,
                             const boost::optional<unsigned char> &fillPattern, const boost::optional<double> &fillFGTransparency,
                             const boost::optional<double> &fillBGTransparency, const boost::optional<unsigned char> &shadowPattern,
-                            const boost::optional<Colour> &shfgc, const boost::optional<double> &shadowOffsetX, const boost::optional<double> &shadowOffsetY);
+                            const boost::optional<Colour> &shfgc, const boost::optional<double> &shadowOffsetX, const boost::optional<double> &shadowOffsetY,
+                            const boost::optional<long> &qsFc, const boost::optional<long> &qsSc, const boost::optional<long> &qsLm) override;
   void collectFillAndShadow(unsigned level, const boost::optional<Colour> &colourFG, const boost::optional<Colour> &colourBG,
                             const boost::optional<unsigned char> &fillPattern, const boost::optional<double> &fillFGTransparency,
                             const boost::optional<double> &fillBGTransparency, const boost::optional<unsigned char> &shadowPattern,
-                            const boost::optional<Colour> &shfgc);
-  void collectThemeReference(unsigned level, const boost::optional<long> &lineColour, const boost::optional<long> &fillColour,
-                             const boost::optional<long> &shadowColour, const boost::optional<long> &fontColour);
-  void collectGeometry(unsigned id, unsigned level, bool noFill, bool noLine, bool noShow);
-  void collectMoveTo(unsigned id, unsigned level, double x, double y);
-  void collectLineTo(unsigned id, unsigned level, double x, double y);
-  void collectArcTo(unsigned id, unsigned level, double x2, double y2, double bow);
+                            const boost::optional<Colour> &shfgc) override;
+  void collectGeometry(unsigned id, unsigned level, bool noFill, bool noLine, bool noShow) override;
+  void collectMoveTo(unsigned id, unsigned level, double x, double y) override;
+  void collectLineTo(unsigned id, unsigned level, double x, double y) override;
+  void collectArcTo(unsigned id, unsigned level, double x2, double y2, double bow) override;
   void collectNURBSTo(unsigned id, unsigned level, double x2, double y2, unsigned char xType, unsigned char yType, unsigned degree,
-                      const std::vector<std::pair<double, double> > &ctrlPnts, const std::vector<double> &kntVec, const std::vector<double> &weights);
-  void collectNURBSTo(unsigned id, unsigned level, double x2, double y2, double knot, double knotPrev, double weight, double weightPrev, unsigned dataID);
-  void collectNURBSTo(unsigned id, unsigned level, double x2, double y2, double knot, double knotPrev, double weight, double weightPrev, const NURBSData &data);
+                      const std::vector<std::pair<double, double> > &ctrlPnts, const std::vector<double> &kntVec, const std::vector<double> &weights) override;
+  void collectNURBSTo(unsigned id, unsigned level, double x2, double y2, double knot, double knotPrev, double weight, double weightPrev, unsigned dataID) override;
+  void collectNURBSTo(unsigned id, unsigned level, double x2, double y2, double knot, double knotPrev, double weight, double weightPrev, const NURBSData &data) override;
   void collectPolylineTo(unsigned id, unsigned level, double x, double y, unsigned char xType, unsigned char yType,
-                         const std::vector<std::pair<double, double> > &points);
-  void collectPolylineTo(unsigned id, unsigned level, double x, double y, unsigned dataID);
-  void collectPolylineTo(unsigned id, unsigned level, double x, double y, const PolylineData &data);
+                         const std::vector<std::pair<double, double> > &points) override;
+  void collectPolylineTo(unsigned id, unsigned level, double x, double y, unsigned dataID) override;
+  void collectPolylineTo(unsigned id, unsigned level, double x, double y, const PolylineData &data) override;
   void collectShapeData(unsigned id, unsigned level, unsigned char xType, unsigned char yType, unsigned degree, double lastKnot,
-                        std::vector<std::pair<double, double> > controlPoints, std::vector<double> knotVector, std::vector<double> weights);
-  void collectShapeData(unsigned id, unsigned level, unsigned char xType, unsigned char yType, std::vector<std::pair<double, double> > points);
-  void collectXFormData(unsigned level, const XForm &xform);
-  void collectTxtXForm(unsigned level, const XForm &txtxform);
-  void collectShapesOrder(unsigned id, unsigned level, const std::vector<unsigned> &shapeIds);
-  void collectForeignDataType(unsigned level, unsigned foreignType, unsigned foreignFormat, double offsetX, double offsetY, double width, double height);
-  void collectPageProps(unsigned id, unsigned level, double pageWidth, double pageHeight, double shadowOffsetX, double shadowOffsetY, double scale);
-  void collectPage(unsigned id, unsigned level, unsigned backgroundPageID, bool isBackgroundPage, const VSDName &pageName);
-  void collectShape(unsigned id, unsigned level, unsigned parent, unsigned masterPage, unsigned masterShape, unsigned lineStyle, unsigned fillStyle, unsigned textStyle);
-  void collectSplineStart(unsigned id, unsigned level, double x, double y, double secondKnot, double firstKnot, double lastKnot, unsigned degree);
-  void collectSplineKnot(unsigned id, unsigned level, double x, double y, double knot);
-  void collectSplineEnd();
-  void collectInfiniteLine(unsigned id, unsigned level, double x1, double y1, double x2, double y2);
-  void collectRelCubBezTo(unsigned id, unsigned level, double x, double y, double a, double b, double c, double d);
-  void collectRelEllipticalArcTo(unsigned id, unsigned level, double x, double y, double a, double b, double c, double d);
-  void collectRelLineTo(unsigned id, unsigned level, double x, double y);
-  void collectRelMoveTo(unsigned id, unsigned level, double x, double y);
-  void collectRelQuadBezTo(unsigned id, unsigned level, double x, double y, double a, double b);
+                        std::vector<std::pair<double, double> > controlPoints, std::vector<double> knotVector, std::vector<double> weights) override;
+  void collectShapeData(unsigned id, unsigned level, unsigned char xType, unsigned char yType, std::vector<std::pair<double, double> > points) override;
+  void collectXFormData(unsigned level, const XForm &xform) override;
+  void collectTxtXForm(unsigned level, const XForm &txtxform) override;
+  void collectShapesOrder(unsigned id, unsigned level, const std::vector<unsigned> &shapeIds) override;
+  void collectForeignDataType(unsigned level, unsigned foreignType, unsigned foreignFormat, double offsetX, double offsetY, double width, double height) override;
+  void collectPageProps(unsigned id, unsigned level, double pageWidth, double pageHeight, double shadowOffsetX, double shadowOffsetY, double scale) override;
+  void collectPage(unsigned id, unsigned level, unsigned backgroundPageID, bool isBackgroundPage, const VSDName &pageName) override;
+  void collectShape(unsigned id, unsigned level, unsigned parent, unsigned masterPage, unsigned masterShape, unsigned lineStyle, unsigned fillStyle, unsigned textStyle) override;
+  void collectSplineStart(unsigned id, unsigned level, double x, double y, double secondKnot, double firstKnot, double lastKnot, unsigned degree) override;
+  void collectSplineKnot(unsigned id, unsigned level, double x, double y, double knot) override;
+  void collectSplineEnd() override;
+  void collectInfiniteLine(unsigned id, unsigned level, double x1, double y1, double x2, double y2) override;
+  void collectRelCubBezTo(unsigned id, unsigned level, double x, double y, double a, double b, double c, double d) override;
+  void collectRelEllipticalArcTo(unsigned id, unsigned level, double x, double y, double a, double b, double c, double d) override;
+  void collectRelLineTo(unsigned id, unsigned level, double x, double y) override;
+  void collectRelMoveTo(unsigned id, unsigned level, double x, double y) override;
+  void collectRelQuadBezTo(unsigned id, unsigned level, double x, double y, double a, double b) override;
 
-  void collectUnhandledChunk(unsigned id, unsigned level);
+  void collectUnhandledChunk(unsigned id, unsigned level) override;
 
-  void collectText(unsigned level, const librevenge::RVNGBinaryData &textStream, TextFormat format);
+  void collectText(unsigned level, const librevenge::RVNGBinaryData &textStream, TextFormat format) override;
   void collectCharIX(unsigned id, unsigned level, unsigned charCount, const boost::optional<VSDName> &font,
                      const boost::optional<Colour> &fontColour, const boost::optional<double> &fontSize, const boost::optional<bool> &bold,
                      const boost::optional<bool> &italic, const boost::optional<bool> &underline, const boost::optional<bool> &doubleunderline,
                      const boost::optional<bool> &strikeout, const boost::optional<bool> &doublestrikeout, const boost::optional<bool> &allcaps,
                      const boost::optional<bool> &initcaps, const boost::optional<bool> &smallcaps, const boost::optional<bool> &superscript,
-                     const boost::optional<bool> &subscript);
+                     const boost::optional<bool> &subscript, const boost::optional<double> &scaleWidth) override;
   void collectDefaultCharStyle(unsigned charCount, const boost::optional<VSDName> &font, const boost::optional<Colour> &fontColour,
                                const boost::optional<double> &fontSize, const boost::optional<bool> &bold, const boost::optional<bool> &italic,
                                const boost::optional<bool> &underline, const boost::optional<bool> &doubleunderline, const boost::optional<bool> &strikeout,
                                const boost::optional<bool> &doublestrikeout, const boost::optional<bool> &allcaps, const boost::optional<bool> &initcaps,
-                               const boost::optional<bool> &smallcaps, const boost::optional<bool> &superscript, const boost::optional<bool> &subscript);
+                               const boost::optional<bool> &smallcaps, const boost::optional<bool> &superscript, const boost::optional<bool> &subscript,
+                               const boost::optional<double> &scaleWidth) override;
   void collectParaIX(unsigned id, unsigned level, unsigned charCount, const boost::optional<double> &indFirst,
                      const boost::optional<double> &indLeft, const boost::optional<double> &indRight, const boost::optional<double> &spLine,
                      const boost::optional<double> &spBefore, const boost::optional<double> &spAfter, const boost::optional<unsigned char> &align,
                      const boost::optional<unsigned char> &bullet, const boost::optional<VSDName> &bulletStr, const boost::optional<VSDName> &bulletFont,
                      const boost::optional<double> &bulletFontSize, const boost::optional<double> &textPosAfterBullet,
-                     const boost::optional<unsigned> &flags);
+                     const boost::optional<unsigned> &flags) override;
   void collectDefaultParaStyle(unsigned charCount, const boost::optional<double> &indFirst, const boost::optional<double> &indLeft,
                                const boost::optional<double> &indRight, const boost::optional<double> &spLine, const boost::optional<double> &spBefore,
                                const boost::optional<double> &spAfter, const boost::optional<unsigned char> &align,
                                const boost::optional<unsigned char> &bullet, const boost::optional<VSDName> &bulletStr,
                                const boost::optional<VSDName> &bulletFont, const boost::optional<double> &bulletFontSize,
-                               const boost::optional<double> &textPosAfterBullet, const boost::optional<unsigned> &flags);
+                               const boost::optional<double> &textPosAfterBullet, const boost::optional<unsigned> &flags) override;
   void collectTextBlock(unsigned level, const boost::optional<double> &leftMargin, const boost::optional<double> &rightMargin,
                         const boost::optional<double> &topMargin, const boost::optional<double> &bottomMargin,
                         const boost::optional<unsigned char> &verticalAlign, const boost::optional<bool> &isBgFilled,
                         const boost::optional<Colour> &bgColour, const boost::optional<double> &defaultTabStop,
-                        const boost::optional<unsigned char> &textDirection);
-  void collectNameList(unsigned id, unsigned level);
-  void collectName(unsigned id, unsigned level,  const librevenge::RVNGBinaryData &name, TextFormat format);
-  void collectPageSheet(unsigned id, unsigned level);
-  void collectMisc(unsigned level, const VSDMisc &misc);
-  void collectLayer(unsigned id, unsigned level, const VSDLayer &layer);
-  void collectLayerMem(unsigned level, const VSDName &layerMem);
-  void collectTabsDataList(unsigned level, const std::map<unsigned, VSDTabSet> &tabSets);
+                        const boost::optional<unsigned char> &textDirection) override;
+  void collectNameList(unsigned id, unsigned level) override;
+  void collectName(unsigned id, unsigned level,  const librevenge::RVNGBinaryData &name, TextFormat format) override;
+  void collectPageSheet(unsigned id, unsigned level) override;
+  void collectMisc(unsigned level, const VSDMisc &misc) override;
+  void collectLayer(unsigned id, unsigned level, const VSDLayer &layer) override;
+  void collectLayerMem(unsigned level, const VSDName &layerMem) override;
+  void collectTabsDataList(unsigned level, const std::map<unsigned, VSDTabSet> &tabSets) override;
 
   // Style collectors
-  void collectStyleSheet(unsigned id, unsigned level, unsigned parentLineStyle, unsigned parentFillStyle, unsigned parentTextStyle);
+  void collectStyleSheet(unsigned id, unsigned level, unsigned parentLineStyle, unsigned parentFillStyle, unsigned parentTextStyle) override;
   void collectLineStyle(unsigned level, const boost::optional<double> &strokeWidth, const boost::optional<Colour> &c, const boost::optional<unsigned char> &linePattern,
                         const boost::optional<unsigned char> &startMarker, const boost::optional<unsigned char> &endMarker,
-                        const boost::optional<unsigned char> &lineCap, const boost::optional<double> &rounding);
+                        const boost::optional<unsigned char> &lineCap, const boost::optional<double> &rounding,
+                        const boost::optional<long> &qsLineColour, const boost::optional<long> &qsLineMatrix) override;
   void collectFillStyle(unsigned level, const boost::optional<Colour> &colourFG, const boost::optional<Colour> &colourBG,
                         const boost::optional<unsigned char> &fillPattern, const boost::optional<double> &fillFGTransparency,
                         const boost::optional<double> &fillBGTransparency, const boost::optional<unsigned char> &shadowPattern,
-                        const boost::optional<Colour> &shfgc, const boost::optional<double> &shadowOffsetX, const boost::optional<double> &shadowOffsetY);
+                        const boost::optional<Colour> &shfgc, const boost::optional<double> &shadowOffsetX, const boost::optional<double> &shadowOffsetY,
+                        const boost::optional<long> &qsFillColour, const boost::optional<long> &qsShadowColour,
+                        const boost::optional<long> &qsFillMatrix) override;
   void collectFillStyle(unsigned level, const boost::optional<Colour> &colourFG, const boost::optional<Colour> &colourBG,
                         const boost::optional<unsigned char> &fillPattern, const boost::optional<double> &fillFGTransparency,
                         const boost::optional<double> &fillBGTransparency, const boost::optional<unsigned char> &shadowPattern,
-                        const boost::optional<Colour> &shfgc);
+                        const boost::optional<Colour> &shfgc) override;
   void collectCharIXStyle(unsigned id, unsigned level, unsigned charCount, const boost::optional<VSDName> &font,
                           const boost::optional<Colour> &fontColour, const boost::optional<double> &fontSize, const boost::optional<bool> &bold,
                           const boost::optional<bool> &italic, const boost::optional<bool> &underline, const boost::optional<bool> &doubleunderline,
                           const boost::optional<bool> &strikeout, const boost::optional<bool> &doublestrikeout, const boost::optional<bool> &allcaps,
                           const boost::optional<bool> &initcaps, const boost::optional<bool> &smallcaps, const boost::optional<bool> &superscript,
-                          const boost::optional<bool> &subscript);
+                          const boost::optional<bool> &subscript, const boost::optional<double> &scaleWidth) override;
   void collectParaIXStyle(unsigned id, unsigned level, unsigned charCount, const boost::optional<double> &indFirst,
                           const boost::optional<double> &indLeft, const boost::optional<double> &indRight, const boost::optional<double> &spLine,
                           const boost::optional<double> &spBefore, const boost::optional<double> &spAfter, const boost::optional<unsigned char> &align,
                           const boost::optional<unsigned char> &bullet, const boost::optional<VSDName> &bulletStr,
                           const boost::optional<VSDName> &bulletFont, const boost::optional<double> &bulletFontSize,
-                          const boost::optional<double> &textPosAfterBullet, const boost::optional<unsigned> &flags);
+                          const boost::optional<double> &textPosAfterBullet, const boost::optional<unsigned> &flags) override;
   void collectTextBlockStyle(unsigned level, const boost::optional<double> &leftMargin, const boost::optional<double> &rightMargin,
                              const boost::optional<double> &topMargin, const boost::optional<double> &bottomMargin,
                              const boost::optional<unsigned char> &verticalAlign, const boost::optional<bool> &isBgFilled,
                              const boost::optional<Colour> &bgColour, const boost::optional<double> &defaultTabStop,
-                             const boost::optional<unsigned char> &textDirection);
-  void collectStyleThemeReference(unsigned level, const boost::optional<long> &lineColour, const boost::optional<long> &fillColour,
-                                  const boost::optional<long> &shadowColour, const boost::optional<long> &fontColour);
+                             const boost::optional<unsigned char> &textDirection) override;
 
-  virtual void collectMetaData(const librevenge::RVNGPropertyList &metaData);
+  void collectMetaData(const librevenge::RVNGPropertyList &metaData) override;
 
 
   // Field list
-  void collectFieldList(unsigned id, unsigned level);
-  void collectTextField(unsigned id, unsigned level, int nameId, int formatStringId);
-  void collectNumericField(unsigned id, unsigned level, unsigned short format, double number, int formatStringId);
+  void collectFieldList(unsigned id, unsigned level) override;
+  void collectTextField(unsigned id, unsigned level, int nameId, int formatStringId) override;
+  void collectNumericField(unsigned id, unsigned level, unsigned short format, unsigned short cellType, double number, int formatStringId) override;
 
-  void startPage(unsigned pageId);
-  void endPage();
-  void endPages();
+  void startPage(unsigned pageId) override;
+  void endPage() override;
+  void endPages() override;
 
 
 private:
@@ -184,14 +184,14 @@ private:
 
   void applyXForm(double &x, double &y, const XForm &xform);
 
-  void transformPoint(double &x, double &y, XForm *txtxform = 0);
-  void transformAngle(double &angle, XForm *txtxform = 0);
+  void transformPoint(double &x, double &y, XForm *txtxform = nullptr);
+  void transformAngle(double &angle, XForm *txtxform = nullptr);
   void transformFlips(bool &flipX, bool &flipY);
 
   double _NURBSBasis(unsigned knot, unsigned degree, double point, const std::vector<double> &knotVector);
 
   void _flushShape();
-  void _flushCurrentPath();
+  void _flushCurrentPath(unsigned id);
   void _flushText();
   void _flushCurrentForeignData();
   void _flushCurrentPage();
@@ -226,6 +226,11 @@ private:
   void _appendVisibleAndPrintable(librevenge::RVNGPropertyList &propList);
   void _bulletFromParaFormat(VSDBullet &bullet, const VSDParaStyle &paraStyle);
   void _listLevelFromBullet(librevenge::RVNGPropertyList &propList, const VSDBullet &bullet);
+  void _fillParagraphProperties(librevenge::RVNGPropertyList &propList, const VSDParaStyle &style);
+  void _fillTabSet(librevenge::RVNGPropertyList &propList, const VSDTabSet &tabSet);
+  void _fillCharProperties(librevenge::RVNGPropertyList &propList, const VSDCharStyle &style);
+  void _convertToPath(const std::vector<librevenge::RVNGPropertyList> &segmentVector,
+                      librevenge::RVNGPropertyListVector &path, double rounding);
 
   bool m_isPageStarted;
   double m_pageWidth;
@@ -238,7 +243,7 @@ private:
   double m_originalX;
   double m_originalY;
   XForm m_xform;
-  XForm *m_txtxform;
+  std::unique_ptr<XForm> m_txtxform;
   VSDMisc m_misc;
   std::vector<librevenge::RVNGPropertyList> m_currentFillGeometry;
   std::vector<librevenge::RVNGPropertyList> m_currentLineGeometry;
@@ -272,12 +277,11 @@ private:
 
   std::map<unsigned, NURBSData> m_NURBSData;
   std::map<unsigned, PolylineData> m_polylineData;
-  librevenge::RVNGBinaryData m_textStream;
+  libvisio::VSDName m_currentText;
   std::map<unsigned, librevenge::RVNGString> m_names, m_stencilNames;
   std::vector<librevenge::RVNGString> m_fields;
   VSDFieldList m_stencilFields;
   unsigned m_fieldIndex;
-  TextFormat m_textFormat;
   std::vector<VSDCharStyle> m_charFormats;
   std::vector<VSDParaStyle> m_paraFormats;
 
@@ -285,8 +289,6 @@ private:
   VSDFillStyle m_fillStyle;
   VSDTextBlockStyle m_textBlockStyle;
 
-  VSDThemeReference m_themeReference;
-
   VSDCharStyle m_defaultCharStyle;
   VSDParaStyle m_defaultParaStyle;
 
@@ -318,6 +320,8 @@ private:
   VSDLayerList m_currentLayerList;
   std::vector<unsigned> m_currentLayerMem;
   std::vector<VSDTabSet> m_tabSets;
+
+  const VSDXTheme *m_documentTheme;
 };
 
 } // namespace libvisio
